/*
 * MyFrameBuffer.cpp
 *
 *  Created on: Aug 13, 2016
 *      Author: nick
 */


#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/videodev2.h>
#include "MyFrameBuffer.h"


void printFix(struct fb_fix_screeninfo finfo)
{
    printf("###################printFix############################\n\n");
    printf("char id[16]; identification string eg TT Builtin: %s\n", finfo.id);
    printf("unsigned long smem_start Start of frame buffer mem (physical address) %lu\n", finfo.smem_start);
    printf("__u32 smem_len; Length of frame buffer mem %du\n", finfo.smem_len);
    printf("__u32 type;see FB_TYPE_* %du\n", finfo.type);
    printf("__u32 type_aux; Interleave for interleaved Planes %du\n", finfo.type_aux);

    printf("__u32 visual; see FB_VISUAL_* %du\n", finfo.visual);
    printf("__u16 xpanstep; zero if no hardware panning %hu\n", finfo.xpanstep);
    printf("__u16 ypanstep; zero if no hardware panning %hu\n", finfo.ypanstep);
    printf("__u16 ywrapstep;  zero if no hardware ywrap %hu\n", finfo.ywrapstep);
    printf("__u32 line_length; length of a line in bytes %du\n", finfo.line_length);
    printf("unsigned long mmio_start; Start of Memory Mapped I/O %lu\n", finfo.mmio_start);
    printf("__u32 mmio_len; Length of Memory Mapped I/O %du\n", finfo.mmio_len);
    printf("__u32 accel; Indicate to driver which specific chip/card we have %du\n", finfo.accel);

    printf("__u16 capabilities;  see FB_CAP_* %hu\n", finfo.capabilities);

}

void printVar(struct fb_var_screeninfo& vinfo)
{
    printf("###################printVar############################\n\n");
    printf("__u32 xres; __u32 yres; visible resolution %du, %du\n",
           vinfo.xres, vinfo.yres);
    printf("__u32 xres_virtual;  __u32 yres_virtual;virtual resolution %du, %du\n",
           vinfo.xres_virtual, vinfo.yres_virtual);

    printf("__u32 xoffset; __u32 yoffset; offset from virtual to visible resolution %du, %du\n",
           vinfo.xoffset, vinfo.yoffset);

    printf("__u32 bits_per_pixel; guess what %du\n", vinfo.bits_per_pixel);
    printf("__u32 grayscale;  0 = color, 1 = grayscale,  >1 = FOURCC %du\n", vinfo.grayscale);
    printf("struct fb_bitfield red; bitfield in fb mem if true color: ");
    printf("__u32 offset;beginning of bitfield__u32 length;length of "
            "bitfield__u32 msb_right; != 0 : Most significant bit is right\n");
    printf("red: offset:%du, length:%du, msb_right:%du\n",
           vinfo.red.offset, vinfo.red.length, vinfo.red.msb_right);
    printf("green: offset:%du, length:%du, msb_right:%du\n",
           vinfo.green.offset, vinfo.green.length, vinfo.green.msb_right);
    printf("blue: offset:%du, length:%du, msb_right:%du\n",
           vinfo.blue.offset, vinfo.blue.length, vinfo.blue.msb_right);
    printf("transp: offset:%du, length:%du, msb_right:%du\n",
           vinfo.transp.offset, vinfo.transp.length, vinfo.transp.msb_right);
    printf("__u32 nonstd;!= 0 Non standard pixel format: %du\n", vinfo.nonstd);

    printf("__u32 activate;  see FB_ACTIVATE_* %du\n", vinfo.activate);

    printf("__u32 height;  height of picture in mm %du\n", vinfo.height);
    printf("__u32 width;   width of picture in mm %du\n", vinfo.width);

    //__u32 accel_flags;      /* (OBSOLETE) see fb_info.flags */

    /* Timing: All values in pixclocks, except pixclock (of course) */
    printf("__u32 pixclock;pixel clock in ps (pico seconds) %du\n", vinfo.pixclock);
    printf("__u32 left_margin; time from sync to picture %du\n", vinfo.left_margin);
    printf("__u32 right_margin; time from picture to sync %du\n", vinfo.right_margin);
    printf("__u32 upper_margin; time from sync to picture %du\n", vinfo.upper_margin);
    printf("__u32 lower_margin; %du\n", vinfo.lower_margin);
    printf("__u32 hsync_len; length of horizontal sync %du\n", vinfo.hsync_len);
    printf("__u32 vsync_len; length of vertical sync %du\n", vinfo.vsync_len);
    printf("__u32 sync;   see FB_SYNC_* %du\n", vinfo.sync);
    printf("__u32 vmode;  see FB_VMODE_* %du\n", vinfo.vmode);
    printf("__u32 rotate; angle we rotate counter clockwise %du\n", vinfo.rotate);
    printf("__u32 colorspace;  colorspace for FOURCC-based modes %du\n", vinfo.colorspace);
//    __u32 reserved[4];      /* Reserved for future compatibility */
}

MyFrameBuffer::MyFrameBuffer()
:m_fbfd(-1), m_fbp(NULL), m_screensize(0)
{

}

bool MyFrameBuffer::init()
{
    m_fbfd = open("/dev/fb0", O_RDWR);
    if (m_fbfd == -1)
    {
        perror("Error: cannot open framebuffer device");
        return false;
    }

    if (ioctl(m_fbfd, FBIOGET_FSCREENINFO, &m_finfo) == -1)
    {
        perror("Error reading fixed information");
        return false;
    }
    if (ioctl(m_fbfd, FBIOGET_VSCREENINFO, &m_vinfo) == -1)
    {
        perror("Error reading variable information");
        return false;
    }
//    m_vinfo.grayscale = V4L2_PIX_FMT_YVU420;
//    m_vinfo.colorspace = V4L2_PIX_FMT_YVU420;
//    m_vinfo.bits_per_pixel = 12;
//    m_vinfo.red.length = 0;
//    m_vinfo.red.offset = 0;
//    m_vinfo.green.length = 0;
//    m_vinfo.green.offset = 0;
//    m_vinfo.blue.length = 0;
//    m_vinfo.blue.offset = 0;
//    if (ioctl(m_fbfd, FBIOPUT_VSCREENINFO, &m_vinfo) == -1)
//    {
//        perror("Error reading variable information");
//        return false;
//    }
    printVar(m_vinfo);
    m_screensize = m_vinfo.xres * m_vinfo.yres * m_vinfo.bits_per_pixel / 8;
    m_fbp = (unsigned char *)mmap(0, m_screensize, PROT_READ | PROT_WRITE, MAP_SHARED, m_fbfd, 0);
    if (m_fbp == MAP_FAILED)
    {
        perror("Error: failed to map framebuffer device to memory");
        return false;
    }
}

MyFrameBuffer::~MyFrameBuffer()
{
    if (m_fbp)
    {
        munmap(m_fbp, m_screensize);
        m_fbp = NULL;
    }
    if (m_fbfd != -1)
    {
        close(m_fbfd);
        m_fbfd = -1;
    }
}

void MyFrameBuffer::draw()
{

    int x = 100, y = 100;       // Where we are going to put the pixel

    // Figure out where in memory to put the pixel
    for (y = 100; y < 300; y++)
    {
        for (x = 100; x < 300; x++)
        {

            int location = (x+m_vinfo.xoffset) * (m_vinfo.bits_per_pixel/8) +
                       (y+m_vinfo.yoffset) * m_finfo.line_length;

            if (m_vinfo.bits_per_pixel == 32) {
                *(m_fbp + location) = 100;        // Some blue
                *(m_fbp + location + 1) = 15+(x-100)/2;     // A little green
                *(m_fbp + location + 2) = 200-(y-100)/5;    // A lot of red
                *(m_fbp + location + 3) = 0;      // No transparency
        //location += 4;
            } else  { //assume 16bpp
                int b = 10;
                int g = (x-100)/6;     // A little green
                int r = 31-(y-100)/16;    // A lot of red
                unsigned short int t = r<<11 | g << 5 | b;
                *((unsigned short int*)(m_fbp + location)) = t;
            }

        }
    }

}
//int main1()
//{
//    int fbfd = 0;
//
//    long int screensize = 0;
//    char *fbp = 0;
//    int x = 0, y = 0;
//    long int location = 0;
//
//    // Open the file for reading and writing
//    fbfd = open("/dev/fb0", O_RDWR);
//    if (fbfd == -1) {
//        perror("Error: cannot open framebuffer device");
//        exit(1);
//    }
//    printf("The framebuffer device was opened successfully.\n");
//
//    // Get fixed screen information
//    if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo) == -1) {
//        perror("Error reading fixed information");
//        exit(2);
//    }
//    printFix(finfo);
//    // Get variable screen information
//    if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo) == -1) {
//        perror("Error reading variable information");
//        exit(3);
//    }
//    printVar(vinfo);
//    vinfo.grayscale = V4L2_PIX_FMT_YVU420;
//    vinfo.colorspace = V4L2_PIX_FMT_YVU420;
//    //vinfo.bits_per_pixel = 0;
//    vinfo.red.length = 0;
//    vinfo.red.offset = 0;
//    vinfo.green.length = 0;
//    vinfo.green.offset = 0;
//    vinfo.blue.length = 0;
//    vinfo.blue.offset = 0;
//
//    //vinfo.nonstd = V4L2_PIX_FMT_YVU420;
//    printf("%dx%d, %dbpp\n", vinfo.xres, vinfo.yres, vinfo.bits_per_pixel);
//
//    if (ioctl(fbfd, FBIOPUT_VSCREENINFO, &vinfo) == -1) {
//        perror("Error reading variable information");
//        exit(3);
//    }
//    printf("**************set var success ***********\n");
//    if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo) == -1) {
//        perror("Error reading variable information");
//        exit(3);
//    }
//    printVar(vinfo);
//    // Figure out the size of the screen in bytes
//    screensize = vinfo.xres * vinfo.yres * vinfo.bits_per_pixel / 8;
//
//    // Map the device to memory
//    fbp = (char *)mmap(0, screensize, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0);
//    if ((int)fbp == -1) {
//        perror("Error: failed to map framebuffer device to memory");
//        exit(4);
//    }
//    printf("The framebuffer device was mapped to memory successfully.\n");
//
//    x = 100; y = 100;       // Where we are going to put the pixel
//
//    // Figure out where in memory to put the pixel
//    for (y = 100; y < 300; y++)
//        for (x = 100; x < 300; x++) {
//
//            location = (x+vinfo.xoffset) * (vinfo.bits_per_pixel/8) +
//                       (y+vinfo.yoffset) * finfo.line_length;
//
//            if (vinfo.bits_per_pixel == 32) {
//                *(fbp + location) = 100;        // Some blue
//                *(fbp + location + 1) = 15+(x-100)/2;     // A little green
//                *(fbp + location + 2) = 200-(y-100)/5;    // A lot of red
//                *(fbp + location + 3) = 0;      // No transparency
//        //location += 4;
//            } else  { //assume 16bpp
//                int b = 10;
//                int g = (x-100)/6;     // A little green
//                int r = 31-(y-100)/16;    // A lot of red
//                unsigned short int t = r<<11 | g << 5 | b;
//                *((unsigned short int*)(fbp + location)) = t;
//            }
//
//        }
//    munmap(fbp, screensize);
//    close(fbfd);
//    return 0;
//}
