/*
 *  V4L2 video capture example, modified by Derek Molloy for the Logitech C920 camera
 *  Modifications, added the -F mode for H264 capture and associated help detail
 *  www.derekmolloy.ie
 *
 *  V4L2 video capture example
 *
 *  This program can be used and distributed without restrictions.
 *
 *      This program is provided with the V4L2 API
 * see http://linuxtv.org/docs.php for more information
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>
#include <getopt.h>             /* getopt_long() */

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
//#include <arm_neon.h>
#include <linux/videodev2.h>

#define CLEAR(x) memset(&(x), 0, sizeof(x))

enum io_method
{
    IO_METHOD_READ, IO_METHOD_MMAP, IO_METHOD_USERPTR,
};

#define CLIP(X) ( (X) > 255 ? 255 : (X) < 0 ? 0 : X)

// RGB -> YUV
#define RGB2Y(R, G, B) CLIP(( (  66 * (R) + 129 * (G) +  25 * (B) + 128) >> 8) +  16)
#define RGB2U(R, G, B) CLIP(( ( -38 * (R) -  74 * (G) + 112 * (B) + 128) >> 8) + 128)
#define RGB2V(R, G, B) CLIP(( ( 112 * (R) -  94 * (G) -  18 * (B) + 128) >> 8) + 128)

// YUV -> RGB
#define C(Y) ( (Y) - 16  )
#define D(U) ( (U) - 128 )
#define E(V) ( (V) - 128 )

#define YUV2R(Y, U, V) CLIP(( 298 * C(Y)              + 409 * E(V) + 128) >> 8)
#define YUV2G(Y, U, V) CLIP(( 298 * C(Y) - 100 * D(U) - 208 * E(V) + 128) >> 8)
#define YUV2B(Y, U, V) CLIP(( 298 * C(Y) + 516 * D(U)              + 128) >> 8)

// RGB -> YCbCr
#define CRGB2Y(R, G, B) CLIP((19595 * R + 38470 * G + 7471 * B ) >> 16)
#define CRGB2Cb(R, G, B) CLIP((36962 * (B - CLIP((19595 * R + 38470 * G + 7471 * B ) >> 16) ) >> 16) + 128)
#define CRGB2Cr(R, G, B) CLIP((46727 * (R - CLIP((19595 * R + 38470 * G + 7471 * B ) >> 16) ) >> 16) + 128)

// YCbCr -> RGB
#define CYCbCr2R(Y, Cb, Cr) CLIP( Y + ( 91881 * Cr >> 16 ) - 179 )
#define CYCbCr2G(Y, Cb, Cr) CLIP( Y - (( 22544 * Cb + 46793 * Cr ) >> 16) + 135)
#define CYCbCr2B(Y, Cb, Cr) CLIP( Y + (116129 * Cb >> 16 ) - 226 )



#define CLIP(color) (unsigned char) (((color) > 0xFF) ? 0xff : (((color) < 0) ? 0 : (color)))

static unsigned int screenWidth = 1920;
static unsigned int screenHeight = 1080;
static unsigned int screenStride = 1920*4;

void convert_yuyv_to_rgb24(const uint8_t *src, uint8_t *dest, int width, int height)
{
        /* From: Hans de Goede <j.w.r.degoede@hhs.nl> */
        int j;

        while(--height >= 0) {
                for(j = 0; j < width; j += 2) {
                        int u = src[1];
                        int v = src[3];
                        int u1 = (((u - 128) << 7) +  (u - 128)) >> 6;
                        int rg = (((u - 128) << 1) +  (u - 128) +
                                  ((v - 128) << 2) + ((v - 128)  << 1)) >> 3;
                        int v1 = (((v - 128) << 1) +  (v - 128)) >> 1;

                        *dest++ = CLIP(src[0] + v1);
                        *dest++ = CLIP(src[0] - rg);
                        *dest++ = CLIP(src[0] + u1);
                        *dest++ = CLIP(src[2] + v1);
                        *dest++ = CLIP(src[2] - rg);
                        *dest++ = CLIP(src[2] + u1);
                        src += 4;
                }
        }
}


#define CLIP(color) (unsigned char) (((color) > 0xFF) ? 0xff : (((color) < 0) ? 0 : (color)))
//void convert_yuyv_to_rgb565(const uint8_t *src, uint16_t *dst, int width, int height)
//{
//	/* From: Hans de Goede <j.w.r.degoede@hhs.nl> */
//
//
//	for (unsigned int i = 0; i < height; i ++)
//	{
//		for(unsigned int j = 0; j < width; j += 2)
//		{
//				int u = src[1];
//				int v = src[3];
//				int u1 = (((u - 128) << 7) +  (u - 128)) >> 6;
//				int rg = (((u - 128) << 1) +  (u - 128) +
//						  ((v - 128) << 2) + ((v - 128)  << 1)) >> 3;
//				int v1 = (((v - 128) << 1) +  (v - 128)) >> 1;
//
//				unsigned short r = 0, g = 0, b = 0;
//				r = CLIP(src[0] + v1) >> 3;
//				g = CLIP(src[0] - rg) >> 2;
//				b = CLIP(src[0] + u1) >> 3;
//
//				dst[0] = (r<<11) | (g<<5) | b;
//				r = CLIP(src[2] + v1) >>3;
//				g = CLIP(src[2] - rg) >>2;
//				b = CLIP(src[2] + u1) >>3;
//				dst[1] = (r<<11) | (g<<5) | b;
//				src += 4;
//				dst += 2;
//		}
//		dst += (1920-width);
//	}
//}

void convert_yuyv_to_rgb565(const uint8_t *src, uint16_t *dst, int width, int height)
{
	/* From: Hans de Goede <j.w.r.degoede@hhs.nl> */


	for (unsigned int i = 0; i < height; i ++)
	{
		for(unsigned int j = 0; j < width; j += 2)
		{
				int u = src[1];
				int v = src[3];
				int u1 = (((u - 128) << 7) +  (u - 128)) >> 6;
				int rg = (((u - 128) << 1) +  (u - 128) +
						  ((v - 128) << 2) + ((v - 128)  << 1)) >> 3;
				int v1 = (((v - 128) << 1) +  (v - 128)) >> 1;

				unsigned short r = 0, g = 0, b = 0;
				r = CLIP(src[0] + v1) >> 3;
				g = CLIP(src[0] - rg) >> 2;
				b = CLIP(src[0] + u1) >> 3;

				dst[0] = (r<<11) | (g<<5) | b;
				r = CLIP(src[2] + v1) >>3;
				g = CLIP(src[2] - rg) >>2;
				b = CLIP(src[2] + u1) >>3;
				dst[1] = (r<<11) | (g<<5) | b;
				src += 4;
				dst += 2;
		}
		dst += (screenWidth-width);
	}
}

void convert_yuyv_to_rgb888(const uint8_t *src, uint8_t *pMap, int width, int height, int stride)
{
	for (unsigned int i = 0; i < height; i ++)
	{
		uint8_t * dst = pMap;
		for(unsigned int j = 0; j < width; j += 2)
		{
			int u = src[1];
			int v = src[3];
			int u1 = (((u - 128) << 7) +  (u - 128)) >> 6;
			int rg = (((u - 128) << 1) +  (u - 128) +
					  ((v - 128) << 2) + ((v - 128)  << 1)) >> 3;
			int v1 = (((v - 128) << 1) +  (v - 128)) >> 1;

			unsigned short r = 0, g = 0, b = 0;
			dst[0] = CLIP(src[0] + v1);
			dst[1] = CLIP(src[0] - rg);
			dst[2] = CLIP(src[0] + u1);
			dst[3] = 0;

			dst[4] = CLIP(src[2] + v1);
			dst[5] = CLIP(src[2] - rg);
			dst[6] = CLIP(src[2] + u1);
			dst[7] = 0;
			src += 4;
			dst += 8;
		}
		pMap += stride;
	}
}
struct buffer
{
    void *start;
    size_t length;
};

static char *dev_name;
static enum io_method io = IO_METHOD_MMAP;
static int fd = -1;
struct buffer *buffers;
static unsigned int n_buffers;
static int out_buf = 0;
static int force_format = 0;
static int frame_count = 100;

static unsigned int width = 640;
static unsigned int height = 480;
static unsigned int stride = 640*4;

static void
errno_exit (const char *s)
{
    fprintf (stderr, "%s error %d, %s\n", s, errno, strerror (errno));
    exit (EXIT_FAILURE);
}

static int
xioctl (int fh, int request, void *arg)
{
    int r;

    do
    {
        r = ioctl (fh, request, arg);
    }
    while (-1 == r && EINTR == errno);

    return r;
}

static void yuv2rgb(unsigned char Y, unsigned char U, unsigned char V, unsigned short* dst)
{
	unsigned short r = 0, g = 0, b = 0;
	r = CYCbCr2R(Y, U, V) << 11;
	g = CYCbCr2G(Y, U, V) << 5;
	b = CYCbCr2B(Y, U, V);
	dst[0] = r | g | b;

}

static void
process_image (const void *p, int size, void* ptr)
{
	if (ptr)
	{
		//memcpy(ptr, p, size);
		//unsigned char* src = (unsigned char*) p;




//		for (unsigned int y = 0; y < height; y ++)
//		{
//			for (unsigned int x = 0; x < (width+1)/2; x +=2)
//			{
//				unsigned char Y1, Y2, U, V;
//				Y1 = src[0];
//				Y2 = src[2];
//				U = src[1];
//				V = src[3];
//				yuv2rgb(Y1, U, V, dst);
//				yuv2rgb(Y2, U, V, dst+1);
//				src += 4;
//				dst += 2;
//			}
//			dst += xWidth;
//		}
		convert_yuyv_to_rgb888((unsigned char*) p, (unsigned char*)ptr, width, height, screenStride);
		//convert_yuyv_to_rgb565(src, dst, width, height);
//		for (unsigned int i=0;i< width*height;i+=2)
//		{
//				unsigned char Y1, Y2, U, V;
//
//				Y1 = src[2*i+0];
//				Y2 = src[2*i+2];
//				U = src[2*i+1];
//				V = src[2*i+3];
//
//				yuv2rgb(Y1, U, V, &dst[3*i]);
//				yuv2rgb(Y2, U, V, &dst[3*(i+1)]);
//			}
	}
	else
	{
		if (out_buf)
		{
			fwrite (p, size, 1, stdout);
		}
	}

    fflush (stderr);
    fprintf (stderr, ".");
    fflush (stdout);
}

static int
read_frame (void* ptr)
{
    struct v4l2_buffer buf;
    unsigned int i;

	CLEAR(buf);

	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_USERPTR;

	if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf))
	{
		switch (errno)
		{
			case EAGAIN:
				return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("VIDIOC_DQBUF");
		}
	}

	for (i = 0; i < n_buffers; ++i)
		if (buf.m.userptr == (unsigned long) buffers[i].start
				&& buf.length == buffers[i].length)
			break;

	assert(i < n_buffers);

	process_image ((void *) buf.m.userptr, buf.bytesused, ptr);

	if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
		errno_exit ("VIDIOC_QBUF");

    return 1;
}

static void
mainloop (void* ptr)
{
    unsigned int count;
    unsigned int loopIsInfinite = 0;

    if (frame_count == 0)
        loopIsInfinite = 1; //infinite loop
    count = frame_count;

    struct timeval time;
    gettimeofday( &time, 0 );

    long start_time = 1000000 * time.tv_sec + time.tv_usec;
    while ((count-- > 0) || loopIsInfinite)
    {
        for (;;)
        {
            fd_set fds;
            struct timeval tv;
            int r;

            FD_ZERO(&fds);
            FD_SET(fd, &fds);

            /* Timeout. */
            tv.tv_sec = 5;
            tv.tv_usec = 0;

            r = select (fd + 1, &fds, NULL, NULL, &tv);

            if (-1 == r)
            {
                if (EINTR == errno)
                    continue;
                errno_exit ("select");
            }

            if (0 == r)
            {
                fprintf (stderr, "select timeout\n");
                exit (EXIT_FAILURE);
            }

            if (read_frame (ptr))
                break;
            /* EAGAIN - continue select loop. */
        }
    }
	long end_time = 1000000 * time.tv_sec + time.tv_usec;
	fprintf(stderr, "total time is %d\n", end_time-start_time);
}

static void
stop_capturing (void)
{
    enum v4l2_buf_type type;


	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (-1 == xioctl (fd, VIDIOC_STREAMOFF, &type))
		errno_exit ("VIDIOC_STREAMOFF");

}

static void
start_capturing (void)
{
    unsigned int i;
    enum v4l2_buf_type type;


	for (i = 0; i < n_buffers; ++i)
	{
		struct v4l2_buffer buf;

		CLEAR(buf);
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_USERPTR;
		buf.index = i;
		buf.m.userptr = (unsigned long) buffers[i].start;
		buf.length = buffers[i].length;

		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			errno_exit ("VIDIOC_QBUF");
	}
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
		errno_exit ("VIDIOC_STREAMON");

}

static void
uninit_device (void)
{
    unsigned int i;


	for (i = 0; i < n_buffers; ++i)
		free (buffers[i].start);


    free (buffers);
}

static void
init_read (unsigned int buffer_size)
{
    buffers = (struct buffer*)calloc (1, sizeof(*buffers));

    if (!buffers)
    {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

    buffers[0].length = buffer_size;
    buffers[0].start = malloc (buffer_size);

    if (!buffers[0].start)
    {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }
}

static void query_buffer(int fd, int n_buffers, struct v4l2_buffer* buf)
{
	CLEAR(*buf);

	buf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf->memory = V4L2_MEMORY_MMAP;
	buf->index = n_buffers;

	if (-1 == xioctl (fd, VIDIOC_QUERYBUF, buf))
	{
		errno_exit ("VIDIOC_QUERYBUF");
	}
	printf("bytesused: %u\n", buf->bytesused);
	printf("flags: %u\n", buf->flags);
	printf("field: %u\n", buf->field);
}

static void
init_mmap (void)
{
    struct v4l2_requestbuffers req;

    CLEAR(req);

    req.count = 4;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;

    if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req))
    {
        if (EINVAL == errno)
        {
            fprintf (stderr, "%s does not support "
                     "memory mapping\n",
                     dev_name);
            exit (EXIT_FAILURE);
        }
        else
        {
            errno_exit ("VIDIOC_REQBUFS");
        }
    }

    if (req.count < 2)
    {
        fprintf (stderr, "Insufficient buffer memory on %s\n", dev_name);
        exit (EXIT_FAILURE);
    }

    buffers = (struct buffer*)calloc (req.count, sizeof(*buffers));

    if (!buffers)
    {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

    for (n_buffers = 0; n_buffers < req.count; ++n_buffers)
    {
    	struct v4l2_buffer buf;

    	query_buffer(fd, n_buffers, &buf);
        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start = mmap (NULL /* start anywhere */, buf.length,
        PROT_READ | PROT_WRITE /* required */,
                                         MAP_SHARED /* recommended */, fd,
                                         buf.m.offset);

        if (MAP_FAILED == buffers[n_buffers].start)
            errno_exit ("mmap");
    }
}

static void
init_userp (unsigned int buffer_size)
{
    struct v4l2_requestbuffers req;

    CLEAR(req);

    req.count = 4;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_USERPTR;

    if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req))
    {
        if (EINVAL == errno)
        {
            fprintf (stderr, "%s does not support "
                     "user pointer i/o\n",
                     dev_name);
            exit (EXIT_FAILURE);
        }
        else
        {
            errno_exit ("VIDIOC_REQBUFS");
        }
    }

    buffers = (struct buffer*) calloc (4, sizeof(*buffers));

    if (!buffers)
    {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

    for (n_buffers = 0; n_buffers < 4; ++n_buffers)
    {
        buffers[n_buffers].length = buffer_size;
        buffers[n_buffers].start = malloc (buffer_size);

        if (!buffers[n_buffers].start)
        {
            fprintf (stderr, "Out of memory\n");
            exit (EXIT_FAILURE);
        }
    }
}

static void
init_device ()
{
    struct v4l2_capability cap;
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;
    struct v4l2_format fmt;
    unsigned int min;

    if (-1 == xioctl (fd, VIDIOC_QUERYCAP, &cap))
    {
        if (EINVAL == errno)
        {
            fprintf (stderr, "%s is no V4L2 device\n", dev_name);
            exit (EXIT_FAILURE);
        }
        else
        {
            errno_exit ("VIDIOC_QUERYCAP");
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE))
    {
        fprintf (stderr, "%s is no video capture device\n", dev_name);
        exit (EXIT_FAILURE);
    }


	if (!(cap.capabilities & V4L2_CAP_STREAMING))
	{
		fprintf (stderr, "%s does not support streaming i/o\n",
				 dev_name);
		exit (EXIT_FAILURE);
	}

    /* Select video input, video standard and tune here. */

    CLEAR(cropcap);

    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (0 == xioctl (fd, VIDIOC_CROPCAP, &cropcap))
    {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect; /* reset to default */

        if (-1 == xioctl (fd, VIDIOC_S_CROP, &crop))
        {
            switch (errno)
            {
                case EINVAL:
                    /* Cropping not supported. */
                    break;
                default:
                    /* Errors ignored. */
                    break;
            }
        }
    }
    else
    {
        /* Errors ignored. */
    }

    CLEAR(fmt);

    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fprintf (stderr, "Force Format %d\n", force_format);


	fmt.fmt.pix.width = width;
	fmt.fmt.pix.height = height;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
	fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

	if (-1 == xioctl (fd, VIDIOC_S_FMT, &fmt))
		errno_exit ("VIDIOC_S_FMT");

        /* Note VIDIOC_S_FMT may change width and height. */

    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 2;
    if (fmt.fmt.pix.bytesperline < min)
        fmt.fmt.pix.bytesperline = min;
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min)
        fmt.fmt.pix.sizeimage = min;


   init_userp (fmt.fmt.pix.sizeimage);

}

static void
close_device (void)
{
    if (-1 == close (fd))
        errno_exit ("close");

    fd = -1;
}

static void
open_device (const char* dev_name)
{
    struct stat st;

    if (-1 == stat (dev_name, &st))
    {
        fprintf (stderr, "Cannot identify '%s': %d, %s\n", dev_name, errno,
                 strerror (errno));
        exit (EXIT_FAILURE);
    }

    if (!S_ISCHR(st.st_mode))
    {
        fprintf (stderr, "%s is no device\n", dev_name);
        exit (EXIT_FAILURE);
    }

    fd = open (dev_name, O_RDWR /* required */| O_NONBLOCK, 0);

    if (-1 == fd)
    {
        fprintf (stderr, "Cannot open '%s': %d, %s\n", dev_name, errno,
                 strerror (errno));
        exit (EXIT_FAILURE);
    }
}

static void
usage (FILE *fp, int argc, char **argv)
{
    fprintf (
            fp,
            "Usage: %s [options]\n\n"
            "Version 1.3\n"
            "Options:\n"
            "-d | --device name   Video device name [%s]\n"
            "-h | --help          Print this message\n"
            "-m | --mmap          Use memory mapped buffers [default]\n"
            "-r | --read          Use read() calls\n"
            "-u | --userp         Use application allocated buffers\n"
            "-o | --output        Outputs stream to stdout\n"
            "-f | --format        Force format to 640x480 YUYV\n"
            "-F | --formatH264    Force format to 1920x1080 H264\n"
            "-c | --count         Number of frames to grab [%i] - use 0 for infinite\n"
            "\n"
            "Example usage: capture -F -o -c 300 > output.raw\n"
            "Captures 300 frames of H264 at 1920x1080 - use raw2mpg4 script to convert to mpg4\n",
            argv[0], dev_name, frame_count);
}

static const char short_options[] = "d:hmruofFc:";

static const struct option long_options[] =
{
{"device", required_argument, NULL, 'd'},
{"help", no_argument, NULL, 'h'},
{"mmap", no_argument, NULL, 'm'},
{"read", no_argument, NULL, 'r'},
{"userp", no_argument, NULL, 'u'},
{"output", no_argument, NULL, 'o'},
{"format", no_argument, NULL, 'f'},
{"formatH264", no_argument, NULL, 'F'},
{"count", required_argument, NULL, 'c'},
{0, 0, 0, 0}};

int
main_func (void* ptr, unsigned int w, unsigned int h, unsigned int stride)
{
    dev_name = "/dev/video0";
    screenWidth = w;
    screenHeight = h;
    screenStride = stride;
    open_device (dev_name);
    init_device ();
    start_capturing ();
    mainloop (ptr);
    stop_capturing ();
    uninit_device ();
    close_device ();
    fprintf (stderr, "\n");
    return 0;
}
